@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2019 Artifex Software, Inc.
+/* Copyright (C) 2001-2023 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
@@ -112,9 +112,9 @@ subst_CID_on_WMode_finalize(const gs_memory_t *cmem, void *data)
     gs_subst_CID_on_WMode_t *subst = (gs_subst_CID_on_WMode_t *)data;
     (void)cmem; /* unused */
 
-    gs_free_object(subst->rc.memory, subst->data + 0, "subst_CID_on_WMode_finalize");
+    gs_free_object(subst->rc.memory, subst->data[0], "subst_CID_on_WMode_finalize");
     subst->data[0] = NULL;
-    gs_free_object(subst->rc.memory, subst->data + 1, "subst_CID_on_WMode_finalize");
+    gs_free_object(subst->rc.memory, subst->data[1], "subst_CID_on_WMode_finalize");
     subst->data[1] = NULL;
 }
 
