@@ -1,14 +1,16 @@
-/* Copyright (C) 2001-2012 Artifex Software, Inc.
+/* Copyright (C) 2001-2024 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
    implied.
 
-   This software is distributed under license and may not be copied, modified
-   or distributed except as expressly authorized under the terms of that
-   license.  Refer to licensing information at http://www.artifex.com/
-   or contact Artifex Software, Inc.,  7 Mt. Lassen Drive - Suite A-134,
-   San Rafael, CA  94903, U.S.A., +1(415)492-9861, for further information.
+   This software is distributed under license and may not be copied,
+   modified or distributed except as expressly authorized under the terms
+   of the license contained in the file LICENSE in this distribution.
+
+   Refer to licensing information at http://www.artifex.com or contact
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 /* Font API support  */
@@ -17,29 +19,13 @@
 #define gxfapi_INCLUDED
 
 #include "gsmemory.h"
-#include "gsmatrix.h"
 #include "gsccode.h"
 #include "stdint_.h"
+#include "gsgstate.h"
+#include "gsfont.h"
+#include "gstext.h"
 
-#ifndef gs_font_base_DEFINED
-#  define gs_font_base_DEFINED
 typedef struct gs_font_base_s gs_font_base;
-#endif
-
-#ifndef gs_gstate_DEFINED
-#  define gs_gstate_DEFINED
-typedef struct gs_gstate_s gs_gstate;
-#endif
-
-#ifndef gs_text_enum_DEFINED
-#  define gs_text_enum_DEFINED
-typedef struct gs_text_enum_s gs_text_enum_t;
-#endif
-
-#ifndef gs_font_DEFINED
-#  define gs_font_DEFINED
-typedef struct gs_font_s gs_font;
-#endif
 
 typedef int fracint;            /* A fractional integer with statically unknown number of fraction bits.
                                    The number of bits depends on plugin and is being specified in
@@ -89,6 +75,36 @@ typedef enum
     gs_fapi_font_feature_BlendAxisTypes,
     gs_fapi_font_feature_BlendPrivate_count,
     gs_fapi_font_feature_BlendFontInfo_count,
+    gs_fapi_font_feature_BlendFontBBox_length,
+    gs_fapi_font_feature_BlendFontBBox,
+
+    gs_fapi_font_feature_BlendBlueValues_length,
+    gs_fapi_font_feature_BlendBlueValues_count,
+    gs_fapi_font_feature_BlendBlueValues,
+    gs_fapi_font_feature_BlendOtherBlues_length,
+    gs_fapi_font_feature_BlendOtherBlues_count,
+    gs_fapi_font_feature_BlendOtherBlues,
+    gs_fapi_font_feature_BlendBlueScale_count,
+    gs_fapi_font_feature_BlendBlueScale,
+    gs_fapi_font_feature_BlendBlueShift_count,
+    gs_fapi_font_feature_BlendBlueShift,
+    gs_fapi_font_feature_BlendBlueFuzz_count,
+    gs_fapi_font_feature_BlendBlueFuzz,
+    gs_fapi_font_feature_BlendForceBold_count,
+    gs_fapi_font_feature_BlendForceBold,
+    gs_fapi_font_feature_BlendStdHW_length,
+    gs_fapi_font_feature_BlendStdHW_count,
+    gs_fapi_font_feature_BlendStdHW,
+    gs_fapi_font_feature_BlendStdVW_length,
+    gs_fapi_font_feature_BlendStdVW_count,
+    gs_fapi_font_feature_BlendStdVW,
+    gs_fapi_font_feature_BlendStemSnapH_length,
+    gs_fapi_font_feature_BlendStemSnapH_count,
+    gs_fapi_font_feature_BlendStemSnapH,
+    gs_fapi_font_feature_BlendStemSnapV_length,
+    gs_fapi_font_feature_BlendStemSnapV_count,
+    gs_fapi_font_feature_BlendStemSnapV,
+
     gs_fapi_font_feature_WeightVector_count,
     gs_fapi_font_feature_WeightVector,
     gs_fapi_font_feature_BlendDesignPositionsArrays_count,
@@ -102,6 +118,12 @@ typedef enum
     /* End CharString emission */
 } gs_fapi_font_feature;
 
+typedef enum
+{
+    gs_fapi_glyph_invalid_format = -1,
+    gs_fapi_glyph_invalid_index = -2
+} gs_fapi_glyph_error;
+
 typedef enum
 {
     gs_fapi_metrics_notdef,
@@ -152,53 +174,60 @@ struct gs_fapi_font_s
     bool is_vertical;
     bool metrics_only;  /* can save the expense of loading entire glyph */
     gs_fapi_ttf_cmap_request ttf_cmap_req[GS_FAPI_NUM_TTF_CMAP_REQ]; /* Lets client request a specific cmap to be set. Also, a couple of fallbacks */
+    gs_fapi_ttf_cmap_request ttf_cmap_selected;                      /* If the request contains multiple cmaps, this records which was actually selected */
     void *client_ctx_p;
     void *client_font_data;
     void *client_font_data2;
     const void *char_data;
     int char_data_len;
     float embolden;
-    unsigned short (*get_word) (gs_fapi_font *ff,
-                                gs_fapi_font_feature var_id, int index);
-    unsigned long (*get_long) (gs_fapi_font *ff, gs_fapi_font_feature var_id,
-                               int index);
-    float (*get_float) (gs_fapi_font *ff, gs_fapi_font_feature var_id,
-                        int index);
-    int (*get_name) (gs_fapi_font *ff, gs_fapi_font_feature var_id,
-                     int index, char *buffer, int len);
-    int (*get_proc) (gs_fapi_font *ff, gs_fapi_font_feature var_id,
-                     int index, char *Buffer);
-    unsigned short (*get_gsubr) (gs_fapi_font *ff, int index, byte *buf,
-                                 ushort buf_length);
-    unsigned short (*get_subr) (gs_fapi_font *ff, int index, byte *buf,
-                                ushort buf_length);
-    unsigned short (*get_raw_subr) (gs_fapi_font *ff, int index, byte *buf,
-                                    ushort buf_length);
-    int (*get_glyph) (gs_fapi_font *ff, int char_code, byte *buf,
-                      ushort buf_length);
-    unsigned short (*serialize_tt_font) (gs_fapi_font *ff, void *buf,
-                                         int buf_size);
-    unsigned short (*get_charstring) (gs_fapi_font *ff, int index,
-                                      byte *buf, ushort buf_length);
-    unsigned short (*get_charstring_name) (gs_fapi_font *ff, int index,
-                                           byte *buf, ushort buf_length);
-    int (*get_glyphdirectory_data) (gs_fapi_font *ff, int char_code,
-                                    const byte **ptr);
-    int (*get_glyphname_or_cid) (gs_font_base *pbfont,
+
+    int (*get_word) (gs_fapi_font *ff, gs_fapi_font_feature var_id, int index, unsigned short *ret);
+
+    int (*get_long) (gs_fapi_font *ff, gs_fapi_font_feature var_id, int index, unsigned long *ret);
+
+    int (*get_float) (gs_fapi_font *ff, gs_fapi_font_feature var_id, int index, float *ret);
+
+    int (*get_name) (gs_fapi_font *ff, gs_fapi_font_feature var_id, int index, char *buffer, int len);
+
+    int (*get_proc) (gs_fapi_font *ff, gs_fapi_font_feature var_id, int index, char *Buffer);
+
+    int (*get_gsubr) (gs_fapi_font *ff, int index, byte *buf, int buf_length);
+
+    int (*get_subr) (gs_fapi_font *ff, int index, byte *buf, int buf_length);
+
+    int (*get_raw_subr) (gs_fapi_font *ff, int index, byte *buf, int buf_length);
+
+    int (*get_glyph) (gs_fapi_font *ff, gs_glyph char_code, byte *buf, int buf_length);
+
+    int (*serialize_tt_font) (gs_fapi_font *ff, void *buf, int buf_size);
+
+    int (*retrieve_tt_font) (gs_fapi_font *ff, void **buf, int *buf_size);
+
+    int (*get_charstring) (gs_fapi_font *ff, int index, byte *buf, ushort buf_length);
+
+    int (*get_charstring_name) (gs_fapi_font *ff, int index, byte *buf, ushort buf_length);
+
+    int (*get_glyphdirectory_data) (gs_fapi_font *ff, int char_code, const byte **ptr);
+
+    int (*get_glyphname_or_cid) (gs_text_enum_t *penum, gs_font_base *pbfont,
                                  gs_string *charstring, gs_string *name,
-                                 int ccode, gs_string *enc_char_name,
-                                 char *font_file_path, gs_fapi_char_ref *cr,
+                                 gs_glyph ccode, gs_string *enc_char_name,
+                                  char *font_file_path, gs_fapi_char_ref *cr,
                                  bool bCID);
-    int (*fapi_get_metrics) (gs_fapi_font *ff, gs_string *char_name,
-                             int cid, double *m, bool vertical);
+
+    int (*fapi_get_metrics) (gs_fapi_font *ff, gs_string *char_name, gs_glyph cid, double *m, bool vertical);
+
     int (*fapi_set_cache) (gs_text_enum_t *penum,
                            const gs_font_base *pbfont,
-                           const gs_string *char_name, int cid,
+                           const gs_string *char_name, gs_glyph cid,
                            const double pwidth[2], const gs_rect *pbbox,
                            const double Metrics2_sbw_default[4],
                            bool *imagenow);
 };
 
+#define GS_FAPI_WEIGHTVECTOR_MAX 16
+
 typedef struct gs_fapi_face_s gs_fapi_face;
 struct gs_fapi_face_s
 {
@@ -207,6 +236,10 @@ struct gs_fapi_face_s
     gs_log2_scale_point log2_scale;
     bool align_to_pixels;
     float HWResolution[2];
+    struct {
+        int count;
+        float values[GS_FAPI_WEIGHTVECTOR_MAX];
+    } WeightVector;
 };
 
 typedef struct gs_fapi_path_s gs_fapi_path;
@@ -247,11 +280,6 @@ typedef struct
     int black_width, black_height;
 } gs_fapi_raster;
 
-#ifndef gs_fapi_server_DEFINED
-#define gs_fapi_server_DEFINED
-typedef struct gs_fapi_server_s gs_fapi_server;
-#endif
-
 typedef enum gs_fapi_descendant_code_s
 {                               /* Possible values are descendant font indices and 4 ones defined below. */
     gs_fapi_descendant_prepared = -1,   /* See FAPI_prepare_font in zfapi.c . */
@@ -305,14 +333,16 @@ struct gs_fapi_server_s
     gs_fapi_font ff;
     int max_bitmap;
     bool use_outline;
+    bool transform_outline;
+    gs_matrix outline_mat;
     uint grid_fit;
     gs_matrix initial_FontMatrix;       /* Font Matrix at the time the font is defined */
     /* Used to use the stored 'OrigFont' entry but */
     /* this did not change f a font was defined    */
     /* using an existing base font */
-         
-         
-         
+
+
+
     gs_fapi_retcode(*ensure_open) (gs_fapi_server *server, const char *param, int param_size);
     gs_fapi_retcode(*get_scaled_font) (gs_fapi_server *server, gs_fapi_font *ff, const gs_fapi_font_scale *scale, const char *xlatmap, gs_fapi_descendant_code dc);
     gs_fapi_retcode(*get_decodingID) (gs_fapi_server *server, gs_fapi_font *ff, const char **decodingID);
@@ -328,10 +358,11 @@ struct gs_fapi_server_s
     gs_fapi_retcode(*get_char_outline_metrics) (gs_fapi_server *server, gs_fapi_font *ff, gs_fapi_char_ref *c, gs_fapi_metrics *metrics);
     gs_fapi_retcode(*get_char_outline) (gs_fapi_server *server, gs_fapi_path *p);
     gs_fapi_retcode(*release_char_data) (gs_fapi_server *server);
-         
+
     gs_fapi_retcode(*release_typeface) (gs_fapi_server *server, void *server_font_data);
     gs_fapi_retcode(*check_cmap_for_GID) (gs_fapi_server *server, uint *index);
     gs_fapi_retcode(*get_font_info) (gs_fapi_server *server, gs_fapi_font *ff, gs_fapi_font_info item, int index, void *data, int *datalen);
+    gs_fapi_retcode(*set_mm_weight_vector) (gs_fapi_server *server, gs_fapi_font *ff, float *wvector, int length);
 
     /*  Some people get confused with terms "font cache" and "character cache".
        "font cache" means a cache for scaled font objects, which mainly
@@ -417,7 +448,7 @@ gs_fapi_do_char(gs_font *pfont, gs_gstate *pgs, gs_text_enum_t *penum, char *fon
  * to hold the requested information - data == NULL, or data_len too
  * small will see data_len set to the required amount, and an error
  * returned.
- */ 
+ */
 int
 gs_fapi_get_font_info(gs_font *pfont, gs_fapi_font_info item, int index,
                       void *data, int *data_len);
@@ -433,4 +464,8 @@ void gs_fapi_finit(gs_memory_t *mem);
 
 
 void gx_fapi_get_ulp_character_data(byte **header, byte **character_data);
+
+void gx_fapi_bits_smear_horizontally(byte *dest, const byte *src, uint width, uint smear_width);
+
+void gx_fapi_bits_merge(byte *dest, const byte *src, uint nbytes);
 #endif /* gxfapi_INCLUDED */
