@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2018 Artifex Software, Inc.
+/* Copyright (C) 2001-2023 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
@@ -59,9 +59,9 @@ static RELOC_PTRS_WITH(stream_reloc_ptrs, stream *st)
             RELOC_VAR(st->cbuf);
         reloc = cbuf_old - st->cbuf;
         /* Relocate the other buffer pointers. */
-        st->srptr -= reloc;
-        st->srlimit -= reloc;	/* same as swptr */
-        st->swlimit -= reloc;
+        st->cursor.r.ptr -= reloc;
+        st->cursor.r.limit -= reloc;	/* same as swptr */
+        st->cursor.w.limit -= reloc;
     }
     RELOC_VAR(st->strm);
     RELOC_VAR(st->prev);
@@ -79,11 +79,11 @@ stream_finalize(const gs_memory_t *cmem, void *vptr)
     stream *const st = vptr;
     (void)cmem; /* unused */
 
-    if_debug2m('u', st->memory, "[u]%s 0x%lx\n",
+    if_debug2m('u', st->memory, "[u]%s "PRI_INTPTR"\n",
                (!s_is_valid(st) ? "already closed:" :
                 st->is_temp ? "is_temp set:" :
                 st->file == 0 ? "not file:" :
-                "closing file:"), (ulong) st);
+                "closing file:"), (intptr_t) st);
     if (s_is_valid(st) && !st->is_temp && st->file != 0) {
         /* Prevent any attempt to free the buffer. */
         st->cbuf = 0;
@@ -110,16 +110,31 @@ s_init(stream *s, gs_memory_t * mem)
     s->prev = s->next = 0;	/* clean for GC */
     s->file_name.data = 0;	/* ibid. */
     s->file_name.size = 0;
+    s->end_status = 0;
+    s->modes = 0;
     s->close_strm = false;	/* default */
     s->close_at_eod = true;	/* default */
+    s->cbuf_string_memory = NULL;
 }
 stream *
 s_alloc(gs_memory_t * mem, client_name_t cname)
 {
     stream *s = gs_alloc_struct(mem, stream, &st_stream, cname);
 
-    if_debug2m('s', mem, "[s]alloc(%s) = 0x%lx\n",
-               client_name_string(cname), (ulong) s);
+    if_debug2m('s', mem, "[s]alloc(%s) = "PRI_INTPTR"\n",
+               client_name_string(cname), (intptr_t) s);
+    if (s == 0)
+        return 0;
+    s_init(s, mem);
+    return s;
+}
+stream *
+s_alloc_immovable(gs_memory_t * mem, client_name_t cname)
+{
+    stream *s = gs_alloc_struct_immovable(mem, stream, &st_stream, cname);
+
+    if_debug2m('s', mem, "[s]alloc(%s) = "PRI_INTPTR"\n",
+               client_name_string(cname), (intptr_t) s);
     if (s == 0)
         return 0;
     s_init(s, mem);
@@ -143,10 +158,10 @@ s_alloc_state(gs_memory_t * mem, gs_memory_type_ptr_t stype,
 {
     stream_state *st = gs_alloc_struct(mem, stream_state, stype, cname);
 
-    if_debug3m('s', mem, "[s]alloc_state %s(%s) = 0x%lx\n",
+    if_debug3m('s', mem, "[s]alloc_state %s(%s) = "PRI_INTPTR"\n",
                client_name_string(cname),
                client_name_string(stype->sname),
-               (ulong) st);
+               (intptr_t) st);
     if (st)
         s_init_state(st, NULL, mem);
     return st;
@@ -159,8 +174,13 @@ s_std_init(register stream * s, byte * ptr, uint len, const stream_procs * pp,
 {
     s->templat = &s_no_template;
     s->cbuf = ptr;
-    s->srptr = s->srlimit = s->swptr = ptr - 1;
-    s->swlimit = ptr - 1 + len;
+
+    /* IMPORTANT: "read" MUST come before "write" - see comment in scommon.h about
+     * the layout of read/write cursor structures.
+     */
+    stream_cursor_read_init(&s->cursor.r, ptr, 0);
+    stream_cursor_write_init(&s->cursor.w, ptr, len);
+
     s->end_status = 0;
     s->foreign = 0;
     s->modes = modes;
@@ -174,9 +194,10 @@ s_std_init(register stream * s, byte * ptr, uint len, const stream_procs * pp,
     s->file = 0;
     s->file_name.data = 0;	/* in case stream is on stack */
     s->file_name.size = 0;
+    s->cbuf_string_memory = NULL;
     if (s->memory) {
-        if_debug4m('s', s->memory, "[s]init 0x%lx, buf=0x%lx, len=%u, modes=%d\n",
-                   (ulong) s, (ulong) ptr, len, modes);
+        if_debug4m('s', s->memory, "[s]init "PRI_INTPTR", buf="PRI_INTPTR", len=%u, modes=%d\n",
+                   (intptr_t) s, (intptr_t) ptr, len, modes);
     }
 }
 
@@ -227,14 +248,14 @@ s_std_null(stream * s)
 void
 s_std_read_reset(stream * s)
 {
-    s->srptr = s->srlimit = s->cbuf - 1;
+    s->cursor.r.ptr = s->cursor.r.limit = s->cbuf - 1;
 }
 
 /* Discard the contents of the buffer when writing. */
 void
 s_std_write_reset(stream * s)
 {
-    s->swptr = s->cbuf - 1;
+    s->cursor.w.ptr = s->cbuf - 1;
 }
 
 /* Flush data to end-of-file when reading. */
@@ -242,7 +263,7 @@ int
 s_std_read_flush(stream * s)
 {
     while (1) {
-        s->srptr = s->srlimit = s->cbuf - 1;
+        s->cursor.r.ptr = s->cursor.r.limit = s->cbuf - 1;
         if (s->end_status)
             break;
         s_process_read_buf(s);
@@ -316,7 +337,7 @@ s_disable(register stream * s)
     }
     /****** SHOULD DO MORE THAN THIS ******/
     if (s->memory) {
-        if_debug1m('s', s->memory, "[s]disable 0x%lx\n", (ulong) s);
+        if_debug1m('s', s->memory, "[s]disable "PRI_INTPTR"\n", (intptr_t) s);
     }
 }
 
@@ -345,7 +366,8 @@ s_filter_close(register stream * s)
 
         if (status != 0 && status != EOFC)
             return status;
-        status = sflush(stemp);
+        if (status != EOFC)
+            status = sflush(stemp);
         if (status != 0 && status != EOFC)
             return status;
     }
@@ -393,7 +415,7 @@ stell(stream * s)
      * The stream might have been closed, but the position
      * is still meaningful in this case.
      */
-    const byte *ptr = (s_is_writing(s) ? s->swptr : s->srptr);
+    const byte *ptr = (s_is_writing(s) ? s->cursor.w.ptr : s->cursor.r.ptr);
 
     return (ptr == 0 ? 0 : ptr + 1 - s->cbuf) + s->position;
 }
@@ -456,7 +478,7 @@ spgetcc(register stream * s, bool close_at_eod)
     int min_left = sbuf_min_left(s);
 
     while (status = s->end_status,
-           left = s->srlimit - s->srptr,
+           left = s->cursor.r.limit - s->cursor.r.ptr,
            left <= min_left && status >= 0
         )
         s_process_read_buf(s);
@@ -474,7 +496,7 @@ spgetcc(register stream * s, bool close_at_eod)
         }
         return status;
     }
-    return *++(s->srptr);
+    return *++(s->cursor.r.ptr);
 }
 
 /* Implementing sputc when the buffer is full, */
@@ -486,7 +508,7 @@ spputc(register stream * s, byte b)
         if (s->end_status)
             return s->end_status;
         if (!sendwp(s)) {
-            *++(s->swptr) = b;
+            *++(s->cursor.w.ptr) = b;
             return b;
         }
         s_process_write_buf(s, false);
@@ -499,9 +521,13 @@ spputc(register stream * s, byte b)
 int
 sungetc(register stream * s, byte c)
 {
-    if (!s_is_reading(s) || s->srptr < s->cbuf || *(s->srptr) != c)
+    /* cbuf == NULL means this stream is stdin, and we shouldn't
+       unread from stdin, ever.
+     */
+    if (s->cbuf == NULL || !s_is_reading(s) ||
+        s->cursor.r.ptr < s->cbuf || *(s->cursor.r.ptr) != c)
         return ERRC;
-    s->srptr--;
+    s->cursor.r.ptr--;
     return 0;
 }
 
@@ -519,10 +545,10 @@ sgets(stream * s, byte * buf, uint nmax, uint * pn)
     while (cw.ptr < cw.limit) {
         int left;
 
-        if ((left = s->srlimit - s->srptr) > min_left) {
-            s->srlimit -= min_left;
+        if ((left = s->cursor.r.limit - s->cursor.r.ptr) > min_left) {
+            s->cursor.r.limit -= min_left;
             stream_move(&s->cursor.r, &cw);
-            s->srlimit += min_left;
+            s->cursor.r.limit += min_left;
         } else {
             uint wanted = cw.limit - cw.ptr;
             int c;
@@ -546,7 +572,7 @@ sgets(stream * s, byte * buf, uint nmax, uint * pn)
                  * update position.  However, we need to reset the read
                  * cursor to indicate that there is no data in the buffer.
                  */
-                s->srptr = s->srlimit = s->cbuf - 1;
+                s->cursor.r.ptr = s->cursor.r.limit = s->cbuf - 1;
                 s->position += cw.ptr - wptr;
                 if (status <= 0 || cw.ptr == cw.limit)
                     break;
@@ -573,13 +599,13 @@ sputs(register stream * s, const byte * str, uint wlen, uint * pn)
 
     if (status >= 0)
         while (len > 0) {
-            uint count = s->swlimit - s->swptr;
+            uint count = s->cursor.w.limit - s->cursor.w.ptr;
 
             if (count > 0) {
                 if (count > len)
                     count = len;
-                memcpy(s->swptr + 1, str, count);
-                s->swptr += count;
+                memcpy(s->cursor.w.ptr + 1, str, count);
+                s->cursor.w.ptr += count;
                 str += count;
                 len -= count;
             } else {
@@ -620,7 +646,7 @@ spskip(register stream * s, gs_offset_t nskip, gs_offset_t *pskipped)
         int status;
 
         n -= sbufavailable(s);
-        s->srptr = s->srlimit;
+        s->cursor.r.ptr = s->cursor.r.limit;
         if (s->end_status) {
             *pskipped = nskip - n;
             return s->end_status;
@@ -633,7 +659,7 @@ spskip(register stream * s, gs_offset_t nskip, gs_offset_t *pskipped)
         --n;
     }
     /* Note that if min_left > 0, n < 0 is possible; this is harmless. */
-    s->srptr += n;
+    s->cursor.r.ptr += n;
     *pskipped = nskip;
     return 0;
 }
@@ -812,14 +838,14 @@ sreadbuf(stream * s, stream_cursor_write * pbuf)
                 eof = strm->end_status == EOFC;
             }
             pw = (prev == 0 ? pbuf : &curr->cursor.w);
-            if_debug4m('s', s->memory, "[s]read process 0x%lx, nr=%u, nw=%u, eof=%d\n",
-                       (ulong) curr, (uint) (pr->limit - pr->ptr),
+            if_debug4m('s', s->memory, "[s]read process "PRI_INTPTR", nr=%u, nw=%u, eof=%d\n",
+                       (intptr_t) curr, (uint) (pr->limit - pr->ptr),
                        (uint) (pw->limit - pw->ptr), eof);
             oldpos = pw->ptr;
             status = (*curr->procs.process) (curr->state, pr, pw, eof);
             pr->limit += left;
-            if_debug5m('s', s->memory, "[s]after read 0x%lx, nr=%u, nw=%u, status=%d, position=%"PRId64"\n",
-                       (ulong) curr, (uint) (pr->limit - pr->ptr),
+            if_debug5m('s', s->memory, "[s]after read "PRI_INTPTR", nr=%u, nw=%u, status=%d, position=%"PRId64"\n",
+                       (intptr_t) curr, (uint) (pr->limit - pr->ptr),
                        (uint) (pw->limit - pw->ptr), status, s->position);
             if (strm == 0 || status != 0)
                 break;
@@ -896,8 +922,8 @@ swritebuf(stream * s, stream_cursor_read * pbuf, bool last)
             else
                 pr = &curr->cursor.r;
             if_debug5m('s', s->memory,
-                       "[s]write process 0x%lx(%s), nr=%u, nw=%u, end=%d\n",
-                       (ulong)curr,
+                       "[s]write process "PRI_INTPTR"(%s), nr=%u, nw=%u, end=%d\n",
+                       (intptr_t)curr,
                        gs_struct_type_name(curr->state->templat->stype),
                        (uint)(pr->limit - pr->ptr),
                        (uint)(pw->limit - pw->ptr), end);
@@ -905,8 +931,8 @@ swritebuf(stream * s, stream_cursor_read * pbuf, bool last)
             if (status >= 0) {
                 status = (*curr->procs.process)(curr->state, pr, pw, end);
                 if_debug5m('s', s->memory,
-                           "[s]after write 0x%lx, nr=%u, nw=%u, end=%d, status=%d\n",
-                           (ulong) curr, (uint) (pr->limit - pr->ptr),
+                           "[s]after write "PRI_INTPTR", nr=%u, nw=%u, end=%d, status=%d\n",
+                           (intptr_t) curr, (uint) (pr->limit - pr->ptr),
                            (uint) (pw->limit - pw->ptr), end, status);
                 if (status == 0 && end)
                     status = EOFC;
@@ -1007,6 +1033,10 @@ static int
                            stream_cursor_write *, bool);
 
 /* Initialize a stream for reading a string. */
+/* String ownership retained by the caller, for example
+   Postscript string objects owned by the Postscript
+   interpreter
+ */
 void
 sread_string(register stream *s, const byte *ptr, uint len)
 {
@@ -1018,22 +1048,49 @@ sread_string(register stream *s, const byte *ptr, uint len)
     s_std_init(s, (byte *)ptr, len, &p, s_mode_read + s_mode_seek);
     s->cbuf_string.data = (byte *)ptr;
     s->cbuf_string.size = len;
+    s->cbuf_string_memory = NULL;
     s->end_status = EOFC;
-    s->srlimit = s->swlimit;
+    s->cursor.r.limit = s->cursor.w.limit;
 }
+
+/* The string ownership is transferred from caller to stream.
+   string_mem pointer must be allocator used to allocate the
+   "string" buffer.
+ */
+void
+sread_transient_string(register stream *s, gs_memory_t *string_mem, const byte *ptr, uint len)
+{
+    static const stream_procs p = {
+         s_string_available, s_string_read_seek, s_std_read_reset,
+         s_std_read_flush, s_std_null, s_string_read_process
+    };
+
+    s_std_init(s, (byte *)ptr, len, &p, s_mode_read + s_mode_seek);
+    s->cbuf_string.data = (byte *)ptr;
+    s->cbuf_string.size = len;
+    s->cbuf_string_memory = string_mem;
+    s->end_status = EOFC;
+    s->cursor.r.limit = s->cursor.w.limit;
+}
+
 /* Initialize a reusable stream for reading a string. */
 static void
 s_string_reusable_reset(stream *s)
 {
-    s->srptr = s->cbuf - 1;	/* just reset to the beginning */
-    s->srlimit = s->srptr + s->bsize;  /* might have gotten reset */
+    s->cursor.r.ptr = s->cbuf - 1;	/* just reset to the beginning */
+    s->cursor.r.limit = s->cursor.r.ptr + s->bsize;  /* might have gotten reset */
 }
 static int
 s_string_reusable_flush(stream *s)
 {
-    s->srptr = s->srlimit = s->cbuf + s->bsize - 1;  /* just set to the end */
+    s->cursor.r.ptr = s->cursor.r.limit = s->cbuf + s->bsize - 1;  /* just set to the end */
     return 0;
 }
+
+/* String ownership retained by the caller, for example
+   Postscript string objects owned by the Postscript
+   interpreter
+ */
 void
 sread_string_reusable(stream *s, const byte *ptr, uint len)
 {
@@ -1051,6 +1108,27 @@ sread_string_reusable(stream *s, const byte *ptr, uint len)
     s->close_at_eod = false;
 }
 
+/* The string ownership is transferred from caller to stream.
+   string_mem pointer must be allocator used to allocate the
+   "string" buffer.
+ */
+void
+sread_transient_string_reusable(stream *s, gs_memory_t *string_mem, const byte *ptr, uint len)
+{
+    /*
+     * Note that s->procs.close is s_close_disable, to parallel
+     * file_close_disable.
+     */
+    static const stream_procs p = {
+         s_string_available, s_string_read_seek, s_string_reusable_reset,
+         s_string_reusable_flush, s_close_disable, s_string_read_process
+    };
+
+    sread_transient_string(s, string_mem, ptr, len);
+    s->procs = p;
+    s->close_at_eod = false;
+}
+
 /* Return the number of available bytes when reading from a string. */
 static int
 s_string_available(stream *s, gs_offset_t *pl)
@@ -1067,9 +1145,9 @@ s_string_read_seek(register stream * s, gs_offset_t pos)
 {
     if (pos < 0 || pos > s->bsize)
         return ERRC;
-    s->srptr = s->cbuf + pos - 1;
+    s->cursor.r.ptr = s->cbuf + pos - 1;
     /* We might be seeking after a reusable string reached EOF. */
-    s->srlimit = s->cbuf + s->bsize - 1;
+    s->cursor.r.limit = s->cbuf + s->bsize - 1;
     /*
      * When the file reaches EOF,
      * stream_compact sets s->position to its end.
@@ -1102,7 +1180,7 @@ s_string_write_seek(register stream * s, gs_offset_t pos)
 {
     if (pos < 0 || pos > s->bsize)
         return ERRC;
-    s->swptr = s->cbuf + pos - 1;
+    s->cursor.w.ptr = s->cbuf + pos - 1;
     return 0;
 }
 
@@ -1206,8 +1284,11 @@ s_add_filter(stream **ps, const stream_template *templat,
     ess->templat = templat;
     ess->memory = mem;
     es->memory = mem;
-    if (s_init_filter(es, ess, buf, bsize, *ps) < 0)
+    if (s_init_filter(es, ess, buf, bsize, *ps) < 0) {
+        gs_free_object(mem, buf, "s_add_filter(buf)");
+        gs_free_object(mem, es, "s_add_filter(stream)");
         return 0;
+    }
     *ps = es;
     return es;
 }
@@ -1219,25 +1300,34 @@ s_add_filter(stream **ps, const stream_template *templat,
 int
 s_close_filters(stream **ps, stream *target)
 {
+    int code = 0;
     while (*ps != target) {
         stream *s = *ps;
         gs_memory_t *mem = s->state->memory;
+        gs_memory_t *cbuf_string_memory = s->cbuf_string_memory;
         byte *sbuf = s->cbuf;
+        byte *cbuf = s->cbuf_string.data;
         stream *next = s->strm;
         int status = sclose(s);
         stream_state *ss = s->state; /* sclose may set this to s */
 
-        if (status < 0)
-            return status;
+        if (code == 0)
+            code = status;
+
+        if (s->cbuf_string_memory != NULL) { /* stream owns string buffer, so free it */
+            gs_free_object(cbuf_string_memory, cbuf, "s_close_filters(cbuf)");
+        }
+
         if (mem) {
-            gs_free_object(mem, sbuf, "s_close_filters(buf)");
+            if (sbuf != cbuf)
+                gs_free_object(mem, sbuf, "s_close_filters(buf)");
             gs_free_object(mem, s, "s_close_filters(stream)");
             if (ss != (stream_state *)s)
                 gs_free_object(mem, ss, "s_close_filters(state)");
         }
         *ps = next;
     }
-    return 0;
+    return code;
 }
 
 /* ------ Stream closing ------ */
