@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2019 Artifex Software, Inc.
+/* Copyright (C) 2001-2024 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
@@ -119,6 +119,8 @@ static const gs_param_item_t pdf_param_items[] = {
     pi("DetectDuplicateImages", gs_param_type_bool, DetectDuplicateImages),
     pi("AllowIncrementalCFF", gs_param_type_bool, AllowIncrementalCFF),
     pi("WantsToUnicode", gs_param_type_bool, WantsToUnicode),
+    pi("WantsOptionalContent", gs_param_type_bool, WantsOptionalContent),
+    pi("PdfmarkCapable", gs_param_type_bool, PdfmarkCapable),
     pi("AllowPSRepeatFunctions", gs_param_type_bool, AllowPSRepeatFunctions),
     pi("IsDistiller", gs_param_type_bool, IsDistiller),
     pi("PreserveSMask", gs_param_type_bool, PreserveSMask),
@@ -128,6 +130,14 @@ static const gs_param_item_t pdf_param_items[] = {
     pi("NoOutputFonts", gs_param_type_bool, FlattenFonts),
     pi("WantsPageLabels", gs_param_type_bool, WantsPageLabels),
     pi("UserUnit", gs_param_type_float, UserUnit),
+    pi("OmitInfoDate", gs_param_type_bool, OmitInfoDate),
+    pi("OmitID", gs_param_type_bool, OmitID),
+    pi("OmitXMP", gs_param_type_bool, OmitXMP),
+    pi("ModifiesPageSize", gs_param_type_bool, ModifiesPageSize),
+    pi("ModifiesPageOrder", gs_param_type_bool, ModifiesPageOrder),
+    pi("WriteObjStms", gs_param_type_bool, WriteObjStms),
+    pi("WriteXRefStm", gs_param_type_bool, WriteXRefStm),
+    pi("ToUnicodeForStdEnc", gs_param_type_bool, ToUnicodeForStdEnc),
 #undef pi
     gs_param_item_end
 };
@@ -254,6 +264,55 @@ gdev_pdf_get_param(gx_device *dev, char *Param, void *list)
             return(param_write_null(plist, "DSC"));
         }
     }
+
+#if OCR_VERSION > 0
+    if (strcmp(Param, "OCRLanguage") == 0) {
+        gs_param_string langstr;
+        if (pdev->ocr_language[0]) {
+            langstr.data = (const byte *)pdev->ocr_language;
+            langstr.size = strlen(pdev->ocr_language);
+            langstr.persistent = false;
+        } else {
+            langstr.data = (const byte *)"eng";
+            langstr.size = 3;
+            langstr.persistent = false;
+        }
+        return param_write_string(plist, "OCRLanguage", &langstr);
+    }
+    if (strcmp(Param, "OCREngine") == 0)
+        return param_write_int(plist, "OCREngine", &pdev->ocr_engine);
+
+    if (strcmp(Param, "UseOCR") == 0) {
+        gs_param_string ocrstr;
+
+        switch(pdev->UseOCR) {
+            case UseOCRNever:
+                ocrstr.data = (const byte *)"Never";
+                ocrstr.size = 5;
+                ocrstr.persistent = false;
+                break;
+            case UseOCRAsNeeded:
+                ocrstr.data = (const byte *)"AsNeeded";
+                ocrstr.size = 8;
+                ocrstr.persistent = false;
+                break;
+            case UseOCRAlways:
+                ocrstr.data = (const byte *)"Always";
+                ocrstr.size = 8;
+                ocrstr.persistent = false;
+                break;
+        }
+        return param_write_string(plist, "UseOCR", &ocrstr);
+    }
+#endif
+
+    if (strcmp(Param, "OmitInfoDate") == 0)
+        return(param_write_bool(plist, "OmitInfoDate", &pdev->OmitInfoDate));
+    if (strcmp(Param, "OmitXMP") == 0)
+        return(param_write_bool(plist, "OmitXMP", &pdev->OmitXMP));
+    if (strcmp(Param, "OmitID") == 0)
+        return(param_write_bool(plist, "OmitID", &pdev->OmitID));
+
     return gdev_psdf_get_param(dev, Param, list);
 }
 
@@ -268,6 +327,49 @@ gdev_pdf_get_params(gx_device * dev, gs_param_list * plist)
     int code;
     int cdv = CoreDistVersion;
 
+#if OCR_VERSION > 0
+    gs_param_string langstr;
+
+    if (pdev->ocr_language[0]) {
+        langstr.data = (const byte *)pdev->ocr_language;
+        langstr.size = strlen(pdev->ocr_language);
+        langstr.persistent = false;
+    } else {
+        langstr.data = (const byte *)"eng";
+        langstr.size = 3;
+        langstr.persistent = false;
+    }
+
+    {
+        gs_param_string ocrstr;
+
+        switch(pdev->UseOCR) {
+            case UseOCRNever:
+                ocrstr.data = (const byte *)"Never";
+                ocrstr.size = 5;
+                ocrstr.persistent = false;
+                break;
+            case UseOCRAsNeeded:
+                ocrstr.data = (const byte *)"AsNeeded";
+                ocrstr.size = 8;
+                ocrstr.persistent = false;
+                break;
+            case UseOCRAlways:
+                ocrstr.data = (const byte *)"Always";
+                ocrstr.size = 8;
+                ocrstr.persistent = false;
+                break;
+        }
+        code = param_write_string(plist, "UseOCR", &ocrstr);
+    }
+    code = param_write_string(plist, "OCRLanguage", &langstr);
+    if(code < 0)
+        return code;
+    code = param_write_int(plist, "OCREngine", &pdev->ocr_engine);
+    if(code < 0)
+        return code;
+#endif
+
     pdev->ParamCompatibilityLevel = cl;
     code = gdev_psdf_get_params(dev, plist);
     if (code < 0 ||
@@ -295,6 +397,7 @@ gdev_pdf_put_params_impl(gx_device * dev, const gx_device_pdf * save_dev, gs_par
     gx_device_pdf *pdev = (gx_device_pdf *) dev;
     float cl = (float)pdev->CompatibilityLevel;
     bool locked = pdev->params.LockDistillerParams, ForOPDFRead;
+    bool XRefStm_set = false, ObjStms_set = false;
     gs_param_name param_name;
 
     pdev->pdf_memory = gs_memory_stable(pdev->memory);
@@ -366,6 +469,89 @@ gdev_pdf_put_params_impl(gx_device * dev, const gx_device_pdf * save_dev, gs_par
         }
     }
 
+#if OCR_VERSION > 0
+    {
+        int len;
+        gs_param_string langstr;
+        switch (code = param_read_string(plist, (param_name = "OCRLanguage"), &langstr)) {
+            case 0:
+                if (gs_is_path_control_active(pdev->memory)
+                && (strlen(pdev->ocr_language) != langstr.size || memcmp(pdev->ocr_language, langstr.data, langstr.size) != 0)) {
+                    return_error(gs_error_invalidaccess);
+                }
+                else {
+                    len = langstr.size;
+                    if (len >= sizeof(pdev->ocr_language))
+                        len = sizeof(pdev->ocr_language)-1;
+                    memcpy(pdev->ocr_language, langstr.data, len);
+                    pdev->ocr_language[len] = 0;
+                }
+                break;
+            case 1:
+                break;
+            default:
+                ecode = code;
+                param_signal_error(plist, param_name, ecode);
+        }
+    }
+
+    {
+        int engine;
+        switch (code = param_read_int(plist, (param_name = "OCREngine"), &engine)) {
+            case 0:
+                pdev->ocr_engine = engine;
+                break;
+            case 1:
+                break;
+            default:
+                ecode = code;
+                param_signal_error(plist, param_name, ecode);
+        }
+    }
+
+    {
+        gs_param_string ocrstr;
+
+        code = param_read_string(plist, (param_name = "UseOCR"), &ocrstr);
+        switch(code) {
+            case 0:
+                if (ocrstr.size == 5 && memcmp(ocrstr.data, "Never", 5) == 0)
+                    pdev->UseOCR = UseOCRNever;
+                if (ocrstr.size == 8 && memcmp(ocrstr.data, "AsNeeded", 8) == 0)
+                    pdev->UseOCR = UseOCRAsNeeded;
+                if (ocrstr.size == 6 && memcmp(ocrstr.data, "Always", 6) == 0)
+                    pdev->UseOCR = UseOCRAlways;
+                break;
+            case 1:
+                break;
+            default:
+                param_signal_error(plist, param_name, code);
+                break;
+        }
+    }
+
+    {
+        gs_param_string ocrstr;
+
+        code = param_read_string(plist, (param_name = "UseOCR"), &ocrstr);
+        switch(code) {
+            case 0:
+                if (ocrstr.size == 5 && memcmp(ocrstr.data, "Never", 5) == 0)
+                    pdev->UseOCR = UseOCRNever;
+                if (ocrstr.size == 8 && memcmp(ocrstr.data, "AsNeeded", 8) == 0)
+                    pdev->UseOCR = UseOCRAsNeeded;
+                if (ocrstr.size == 6 && memcmp(ocrstr.data, "Always", 6) == 0)
+                    pdev->UseOCR = UseOCRAlways;
+                break;
+            case 1:
+                break;
+            default:
+                param_signal_error(plist, param_name, code);
+                break;
+        }
+    }
+#endif
+
     /*
      * Check for LockDistillerParams before doing anything else.
      * If LockDistillerParams is true and is not being set to false,
@@ -432,28 +618,25 @@ gdev_pdf_put_params_impl(gx_device * dev, const gx_device_pdf * save_dev, gs_par
         case 1:
             break;
     }
-    {   /* HACK : gs_param_list_s::memory is documented in gsparam.h as
-           "for allocating coerced arrays". Not sure why zputdeviceparams
-           sets it to the current memory space, while the device
-           assumes to store them in the device's memory space.
-           As a hackish workaround we temporary replace it here.
-           Doing so because we don't want to change the global code now
-           because we're unable to test it with all devices.
-           Bug 688531 "Segmentation fault running pdfwrite from 219-01.ps".
 
-           This solution to be reconsidered after fixing
-           the bug 688533 "zputdeviceparams specifies a wrong memory space.".
-        */
-        gs_memory_t *mem = plist->memory;
+    ecode = param_read_bool(plist, (param_name = "WriteXRefStm"), &pdev->WriteXRefStm);
+    if (ecode < 0)
+        param_signal_error(plist, param_name, ecode);
+    if (ecode == 0)
+        XRefStm_set = true;
+    ecode = param_read_bool(plist, (param_name = "WriteObjStms"), &pdev->WriteObjStms);
+    if (ecode < 0)
+        param_signal_error(plist, param_name, ecode);
+    if (ecode == 0)
+        ObjStms_set = true;
 
-        plist->memory = pdev->pdf_memory;
-        code = gs_param_read_items(plist, pdev, pdf_param_items);
+    {
+        code = gs_param_read_items(plist, pdev, pdf_param_items, pdev->pdf_memory);
         if (code < 0 || (code = param_read_bool(plist, "ForOPDFRead", &ForOPDFRead)) < 0)
         {
         }
         if (code == 0 && !pdev->is_ps2write && !(locked && pdev->params.LockDistillerParams))
             pdev->ForOPDFRead = ForOPDFRead;
-        plist->memory = mem;
     }
     if (code < 0)
         ecode = code;
@@ -463,7 +646,7 @@ gdev_pdf_put_params_impl(gx_device * dev, const gx_device_pdf * save_dev, gs_par
          * has just been opened and nothing has been written,
          * or if we are setting it to the same value.
          */
-        long fon = pdev->FirstObjectNumber;
+        int64_t fon = pdev->FirstObjectNumber;
 
         if (fon != save_dev->FirstObjectNumber) {
             if (fon <= 0 || fon > 0x7fff0000 ||
@@ -501,6 +684,25 @@ gdev_pdf_put_params_impl(gx_device * dev, const gx_device_pdf * save_dev, gs_par
         param_signal_error(plist, param_name, code);
     }
 
+    if (pdev->OmitInfoDate && pdev->PDFX != 0) {
+        emprintf(pdev->memory, "\nIt is not possible to omit the CreationDate when creating PDF/X\nOmitInfoDate is being ignored.\n");
+        pdev->OmitInfoDate = 0;
+    }
+
+    if (pdev->OmitID && pdev->CompatibilityLevel > 1.7) {
+        emprintf(pdev->memory, "\nIt is not possible to omit the ID array when creating a version 2.0 or greater PDF\nOmitID is being ignored.\n");
+        pdev->OmitID = 0;
+    }
+    if (pdev->OmitID && pdev->OwnerPassword.size != 0) {
+        emprintf(pdev->memory, "\nIt is not possible to omit the ID array when creating an encrypted PDF\nOmitID is being ignored.\n");
+        pdev->OmitID = 0;
+    }
+
+    if (pdev->OmitXMP && pdev->PDFA != 0) {
+        emprintf(pdev->memory, "\nIt is not possible to omit the XMP metadta when creating a PDF/A\nOmitXMP is being ignored.\n");
+        pdev->OmitXMP = 0;
+    }
+
     /* PDFA and PDFX are stored in the page device dictionary and therefore
      * set on every setpagedevice. However, if we have encountered a file which
      * can't be made this way, and the PDFACompatibilityPolicy is 1, we want to
@@ -535,17 +737,63 @@ gdev_pdf_put_params_impl(gx_device * dev, const gx_device_pdf * save_dev, gs_par
     if (pdev->PDFA == 1 || pdev->PDFX || pdev->CompatibilityLevel < 1.4) {
          pdev->HaveTransparency = false;
          pdev->PreserveSMask = false;
+         pdev->WantsOptionalContent = false;
     }
 
+    /* We cannot guarantee that all page objects have a Group with a /CS (ColorSpace) entry
+     * which is a requirement for PDF/A-2+ when objects are defined in Device Independent Colour.
+     * Try and warn the user.
+     */
+    if (pdev->PDFA > 1 && pdev->params.ColorConversionStrategy == ccs_UseDeviceIndependentColor) {
+        switch (pdev->PDFACompatibilityPolicy) {
+            case 0:
+                emprintf(pdev->memory,
+                 "\n\tpdfwrite cannot guarantee creating a conformant PDF/A-2 file with device-independent colour.\n\tWe recommend converting to a device colour space.\n\tReverting to normal output.\n");
+                pdev->AbortPDFAX = true;
+                pdev->PDFX = 0;
+                break;
+            case 1:
+                emprintf(pdev->memory,
+                 "\n\tpdfwrite cannot guarantee creating a conformant PDF/A-2 file with device-independent colour.\n\tWe recommend converting to a device colour space.\n\tWe cannot ignore this request, reverting to normal output.\n");
+                break;
+            case 2:
+                emprintf(pdev->memory,
+                 "\n\tpdfwrite cannot guarantee creating a conformant PDF/A-2 file with device-independent colour.\n\tWe recommend converting to a device colour space.\n\tAborting.\n");
+                return_error(gs_error_unknownerror);
+                break;
+            default:
+                emprintf(pdev->memory,
+                 "\n\tpdfwrite cannot guarantee creating a conformant PDF/A-2 file with device-independent colour.\n\tWe recommend converting to a device colour space.\n\tReverting to normal output.\n");
+                pdev->AbortPDFAX = true;
+                pdev->PDFX = 0;
+                break;
+        }
+    }
     /*
      * We have to set version to the new value, because the set of
      * legal parameter values for psdf_put_params varies according to
      * the version.
      */
-    if (pdev->PDFX)
+    if (pdev->PDFX) {
         cl = (float)1.3; /* Instead pdev->CompatibilityLevel = 1.2; - see below. */
-    if (pdev->PDFA == 1 && cl != 1.4)
+        if (pdev->WriteObjStms && ObjStms_set)
+            emprintf(pdev->memory, "Can't use ObjStm before PDF 1.5, PDF/X does not support PDF 1.5, ignoring WriteObjStms directive\n");
+        if (pdev->WriteXRefStm && XRefStm_set)
+            emprintf(pdev->memory, "Can't use an XRef stream before PDF 1.5, PDF/X does not support PDF 1.5, ignoring WriteXRefStm directive\n");
+
+        pdev->WriteObjStms = false;
+        pdev->WriteXRefStm = false;
+    }
+    if (pdev->PDFA == 1 && cl != 1.4) {
         cl = (float)1.4;
+
+        if (pdev->WriteObjStms && ObjStms_set)
+            emprintf(pdev->memory, "Can't use ObjStm before PDF 1.5, PDF/A-1 does not support PDF 1.5, ignoring WriteObjStms directive\n");
+        if (pdev->WriteXRefStm && XRefStm_set)
+            emprintf(pdev->memory, "Can't use an XRef stream before PDF 1.5, PDF/A-1 does not support PDF 1.5, ignoring WriteXRefStm directive\n");
+        pdev->WriteObjStms = false;
+        pdev->WriteXRefStm = false;
+    }
     if (pdev->PDFA == 2 && cl < 1.7)
         cl = (float)1.7;
     pdev->version = (cl < 1.2 ? psdf_version_level2 : psdf_version_ll3);
@@ -568,6 +816,8 @@ gdev_pdf_put_params_impl(gx_device * dev, const gx_device_pdf * save_dev, gs_par
     if (cl < 1.2) {
         pdev->HaveCFF = false;
     }
+    if (cl < 1.5)
+        pdev->WantsOptionalContent = false;
 
     ecode = param_read_float(plist, "UserUnit", &pdev->UserUnit);
     if (ecode < 0)
@@ -594,9 +844,6 @@ gdev_pdf_put_params_impl(gx_device * dev, const gx_device_pdf * save_dev, gs_par
             if (pdev->params.ColorConversionStrategy == ccs_Gray) {
                 emprintf(pdev->memory, "ConvertCMYKImagesToRGB is not compatible with ColorConversionStrategy of Gray\n");
             } else {
-                if (pdev->icc_struct)
-                    rc_decrement(pdev->icc_struct,
-                                 "reset default profile\n");
                 pdf_set_process_color_model(pdev,1);
                 ecode = gsicc_init_device_profile_struct((gx_device *)pdev, NULL, 0);
                 if (ecode < 0)
@@ -604,6 +851,16 @@ gdev_pdf_put_params_impl(gx_device * dev, const gx_device_pdf * save_dev, gs_par
             }
         }
     }
+    /* Record if the user has specified a custom profile, so we don't replace it
+       with a default when we change color space - see win_pr2_set_bpp()
+       Here, we only want to know *if* we have a user specced profile, actually
+       setting it will be handled by gdev_vector_put_params()
+     */
+    if (!pdev->user_icc) {
+        gs_param_string icc_pro_dummy;
+
+        pdev->user_icc = param_read_string(plist, "OutputICCProfile", &icc_pro_dummy) == 0;
+    }
     switch (pdev->params.ColorConversionStrategy) {
         case ccs_ByObjectType:
         case ccs_LeaveColorUnchanged:
@@ -613,34 +870,31 @@ gdev_pdf_put_params_impl(gx_device * dev, const gx_device_pdf * save_dev, gs_par
             pdev->params.TransferFunctionInfo = tfi_Apply;
             break;
         case ccs_CMYK:
-            if (pdev->icc_struct)
-                rc_decrement(pdev->icc_struct,
-                             "reset default profile\n");
             pdf_set_process_color_model(pdev, 2);
-            ecode = gsicc_init_device_profile_struct((gx_device *)pdev, NULL, 0);
-            if (ecode < 0)
-                goto fail;
+            if (!pdev->user_icc) {
+                ecode = gsicc_init_device_profile_struct((gx_device *)pdev, NULL, 0);
+                if (ecode < 0)
+                    goto fail;
+            }
             break;
         case ccs_Gray:
-            if (pdev->icc_struct)
-                rc_decrement(pdev->icc_struct,
-                             "reset default profile\n");
             pdf_set_process_color_model(pdev,0);
-            ecode = gsicc_init_device_profile_struct((gx_device *)pdev, NULL, 0);
-            if (ecode < 0)
-                goto fail;
+            if (!pdev->user_icc) {
+                ecode = gsicc_init_device_profile_struct((gx_device *)pdev, NULL, 0);
+                if (ecode < 0)
+                    goto fail;
+            }
             break;
         case ccs_sRGB:
         case ccs_RGB:
             /* Only bother to do this if we didn't handle it above */
             if (!pdev->params.ConvertCMYKImagesToRGB) {
-                if (pdev->icc_struct)
-                    rc_decrement(pdev->icc_struct,
-                                 "reset default profile\n");
                 pdf_set_process_color_model(pdev,1);
-                ecode = gsicc_init_device_profile_struct((gx_device *)pdev, NULL, 0);
-                if (ecode < 0)
-                    goto fail;
+                if (!pdev->user_icc) {
+                    ecode = gsicc_init_device_profile_struct((gx_device *)pdev, NULL, 0);
+                    if (ecode < 0)
+                        goto fail;
+                }
             }
             break;
         default:
@@ -701,6 +955,11 @@ gdev_pdf_put_params_impl(gx_device * dev, const gx_device_pdf * save_dev, gs_par
         }
     }
 
+    if (pdev->Linearise && pdev->file != NULL && !gp_fseekable(pdev->file)) {
+        emprintf(pdev->memory, "Can't linearise a non-seekable output file, ignoring\n");
+        pdev->Linearise = false;
+    }
+
     if (pdev->Linearise && pdev->is_ps2write) {
         emprintf(pdev->memory, "Can't linearise PostScript output, ignoring\n");
         pdev->Linearise = false;
@@ -711,6 +970,44 @@ gdev_pdf_put_params_impl(gx_device * dev, const gx_device_pdf * save_dev, gs_par
         pdev->Linearise = false;
     }
 
+    /* Object streams and XRef streams */
+    if (pdev->is_ps2write) {
+        pdev->WriteObjStms = false;
+        pdev->WriteXRefStm = false;
+    } else {
+        if (pdev->Linearise) {
+            if (XRefStm_set)
+                emprintf(pdev->memory, "We don't support XRefStm with FastWebView (Linearized PDF), ignoring WriteXRefStm directive\n");
+            pdev->WriteXRefStm = false;
+            if (ObjStms_set)
+                emprintf(pdev->memory, "We don't support ObjStms with FastWebView (Linearized PDF), ignoring WriteObjStms directive\n");
+            pdev->WriteObjStms = false;
+        }
+    }
+    if (pdev->WriteObjStms && pdev->CompatibilityLevel < 1.5) {
+        if (ObjStms_set)
+            emprintf(pdev->memory, "Can't use Object streams before PDF 1.5, ignoring WriteObjStms directive\n");
+        pdev->WriteObjStms = false;
+    }
+    if (pdev->WriteXRefStm && pdev->CompatibilityLevel < 1.5) {
+        if (XRefStm_set)
+            emprintf(pdev->memory, "Can't use an XRef stream before PDF 1.5, ignoring WriteXRefStm directive\n");
+        pdev->WriteXRefStm = false;
+    }
+    if (pdev->WriteObjStms && !pdev->WriteXRefStm) {
+        if (ObjStms_set)
+            emprintf(pdev->memory, "Can't use Object streams without XRef stream, ignoring WriteObjStms directive\n");
+        pdev->WriteObjStms = false;
+    }
+    if (pdev->WriteObjStms) {
+        if (pdev->next_id == 1)
+            pdev->doubleXref = true;
+        if (!pdev->doubleXref) {
+            emprintf(pdev->memory, "Already started writing output, too late to select WriteObjStms, ignoring directive\n");
+            pdev->WriteObjStms = false;
+        }
+    }
+
     if (pdev->FlattenFonts)
         pdev->PreserveTrMode = false;
     return 0;
