@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2012 Artifex Software, Inc.
+/* Copyright (C) 2001-2024 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  7 Mt. Lassen Drive - Suite A-134, San Rafael,
-   CA  94903, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
@@ -56,14 +56,14 @@ pdf_make_soft_mask_dict(gx_device_pdf * pdev, const gs_pdf14trans_params_t * ppa
         if (code < 0)
             return code;
     }
-    if (pparams->transfer_function != NULL) {
-        long id;
+    if (pdev->CompatibilityLevel <= 1.7 && pparams->transfer_function != NULL && pdev->params.TransferFunctionInfo == tfi_Preserve) {
+        int64_t id;
         char buf[20];
 
         code = pdf_write_function(pdev, pparams->transfer_function, &id);
         if (code < 0)
             return code;
-        gs_sprintf(buf, " %ld 0 R", id);
+        gs_snprintf(buf, sizeof(buf), " %ld 0 R", id);
         code = cos_dict_put_c_key_string(soft_mask_dict, "/TR", (const byte *)buf, strlen(buf));
         if (code < 0)
             return code;
@@ -107,11 +107,15 @@ pdf_make_group_dict(gx_device_pdf * pdev, const gs_pdf14trans_params_t * pparams
        a group color specified.
        In this case, the parent group is inherited from
        the previous group or the device color space */
-    if (pgs != NULL && pparams->group_color != UNKNOWN) {
-        const gs_color_space *cs = gs_currentcolorspace_inline(pgs);
+    if (pgs != NULL && pparams->ColorSpace != NULL) {
+        const gs_color_space *cs = pparams->ColorSpace;
 
-        code = pdf_color_space_named(pdev, pgs, &cs_value, NULL, cs,
-                &pdf_color_space_names, false, NULL, 0, false);
+        if (pparams->ColorSpace == NULL)
+            code = pdf_color_space_named(pdev, pgs, &cs_value, NULL, cs,
+                    &pdf_color_space_names, false, NULL, 0, false);
+        else
+            code = pdf_color_space_named(pdev, pgs, &cs_value, NULL, pparams->ColorSpace,
+                    &pdf_color_space_names, false, NULL, 0, false);
         if (code < 0)
             return code;
         code = cos_dict_put_c_key(group_dict, "/CS", &cs_value);
@@ -162,15 +166,64 @@ pdf_make_form_dict(gx_device_pdf * pdev, const gs_pdf14trans_params_t * pparams,
     code = cos_dict_put_c_key_object(form_dict, "/BBox", (cos_object_t *)bbox_array);
     if (code < 0)
         return code;
+    if (pdev->PendingOC != 0) {
+        if (pdev->CompatibilityLevel < 1.4999) {
+            if (pdev->PDFA) {
+                switch (pdev->PDFACompatibilityPolicy) {
+                    case 0:
+                        emprintf(pdev->memory,
+                                 "Optional Content not valid in this version of PDF, reverting to normal PDF output\n");
+                        pdev->AbortPDFAX = true;
+                        pdev->PDFA = 0;
+                        break;
+                    case 1:
+                        emprintf(pdev->memory,
+                                 "Optional Content not valid in this version of PDF. Dropping feature to preserve PDF/A compatibility\n");
+                        break;
+                    case 2:
+                        emprintf(pdev->memory,
+                                 "Optional Content not valid in this version of PDF,  aborting conversion\n");
+                        return_error (gs_error_typecheck);
+                        break;
+                    default:
+                        emprintf(pdev->memory,
+                                 "Optional Content not valid in this version of PDF, unrecognised PDFACompatibilityLevel,\nreverting to normal PDF output\n");
+                        pdev->AbortPDFAX = true;
+                        pdev->PDFA = 0;
+                        break;
+                }
+            } else {
+                emprintf(pdev->memory,
+                         "Optional Content not valid in this version of PDF. Dropping feature to preserve compatibility\n");
+            }
+        } else {
+            char str[256];
+            gs_param_string param;
+            cos_object_t *pco = NULL;
+
+            gs_snprintf(str, sizeof(str), "{Obj%dG0}", pdev->PendingOC);
+            param.data = (const byte *)str;
+            param.size = strlen(str);
+            code = pdf_refer_named(pdev, &param, &pco);
+            if(code < 0)
+                return code;
+
+            gs_snprintf(str, sizeof(str), "%ld 0 R", pco->id);
+            code = cos_dict_put_string_copy(form_dict, "/OC", str);
+            if (code < 0)
+                return code;
+
+            pdev->PendingOC = 0;
+        }
+    }
     return cos_dict_put_c_key_object(form_dict, "/Group", (cos_object_t *)group_dict);
 }
 
 static int
 pdf_begin_transparency_group(gs_gstate * pgs, gx_device_pdf * pdev,
-                                const gs_pdf14trans_params_t * pparams)
+                                const gs_pdf14trans_params_t * pparams, bool page_group)
 {
     cos_dict_t *group_dict;
-    bool in_page = is_in_page(pdev);
     int code;
 
     if (pgs == NULL)
@@ -182,23 +235,26 @@ pdf_begin_transparency_group(gs_gstate * pgs, gx_device_pdf * pdev,
     if (code < 0)
         return code;
     code = pdf_check_soft_mask(pdev, pgs);
+    if (code < 0)
+        return code;
     if (pdf_must_put_clip_path(pdev, pgs->clip_path)) {
         code = pdf_put_clip_path(pdev, pgs->clip_path);
         if (code < 0)
             return code;
     }
-    if (!in_page)
+    if (page_group)
         pdev->pages[pdev->next_page].group_id = group_dict->id;
     else if (pparams->image_with_SMask) {
         /* An internal group for the image implementation.
            See doimagesmask in gs/lib/pdf_draw.ps .
            Just set a flag for skipping pdf_end_transparency_group. */
         pdev->image_with_SMask |= 1 << ++pdev->FormDepth;
+        pdev->PatternsSinceForm = 0;
     } else {
         pdf_resource_t *pres, *pres_gstate = NULL;
         cos_dict_t *pcd = NULL, *pcd_Resources = NULL;
 
-        code = pdf_prepare_drawing(pdev, pgs, &pres_gstate);
+        code = pdf_prepare_drawing(pdev, pgs, &pres_gstate, false);
         if (code < 0)
             return code;
         code = pdf_end_gstate(pdev, pres_gstate);
@@ -209,6 +265,7 @@ pdf_begin_transparency_group(gs_gstate * pgs, gx_device_pdf * pdev,
         if (code < 0)
             return code;
         pdev->FormDepth++;
+        pdev->PatternsSinceForm = 0;
         code = pdf_make_form_dict(pdev, pparams, pgs, group_dict, (cos_dict_t *)pres->object);
         if (code < 0)
             return code;
@@ -230,12 +287,13 @@ pdf_end_transparency_group(gs_gstate * pgs, gx_device_pdf * pdev)
 {
     int bottom = (pdev->ResourcesBeforeUsage ? 1 : 0);
 
-    if (!is_in_page(pdev))
-        return 0;	/* corresponds to check in pdf_begin_transparency_group */
+    if (!is_in_page(pdev) && pdev->sbstack_depth == 0)
+        return 0;	/* A Group definition at the page level, handled separately. */
     if (pdev->image_with_SMask & (1 << pdev->FormDepth)) {
         /* An internal group for the image implementation.
            See pdf_begin_transparency_group. */
         pdev->image_with_SMask &= ~(1 << pdev->FormDepth--);
+        pdev->PatternsSinceForm = 0;
         return 0;
     } else if (pdev->sbstack_depth == bottom) {
         /* We're closing the page group. */
@@ -248,6 +306,7 @@ pdf_end_transparency_group(gs_gstate * pgs, gx_device_pdf * pdev)
         uint ignore;
 
         pdev->FormDepth--;
+        pdev->PatternsSinceForm = 0;
         code = pdf_exit_substream(pdev);
         if (code < 0)
             return code;
@@ -269,7 +328,7 @@ pdf_begin_transparency_mask(gs_gstate * pgs, gx_device_pdf * pdev,
                                 const gs_pdf14trans_params_t * pparams)
 {
     if (pparams->subtype == TRANSPARENCY_MASK_None) {
-        int code, id = pgs->soft_mask_id;
+        int code;
         pdf_resource_t *pres = 0L;
 
         /* reset the soft mask ID. Apparently this is only used by pdfwrite, if we don't
@@ -277,12 +336,13 @@ pdf_begin_transparency_mask(gs_gstate * pgs, gx_device_pdf * pdev,
          * changed, and so doesn't write out the GState
          */
         pgs->soft_mask_id = 0;
-        code = pdf_prepare_drawing(pdev, pgs, &pres);
+        code = pdf_prepare_drawing(pdev, pgs, &pres, false);
         if (code == gs_error_interrupt) {
-            /* Not in an appropriate context, ignore it but restore
-             * the old soft_mask_id. Not sure this is correct, but it works for now.
+            /* */
+            /* Not in an appropriate context.  Do not restore the soft_mask_id.
+               Otherwise any group push that occurs following this will use that
+               softmask, which clearly should be NONE here.
              */
-            pgs->soft_mask_id = id;
             /* ignore return code, we don't care about this graphics state as we aren't
              * emitting it anyway
              */
@@ -328,13 +388,14 @@ pdf_begin_transparency_mask(gs_gstate * pgs, gx_device_pdf * pdev,
     } else {
         int code;
 
+        pdev->smask_construction = true;
         code = pdf_make_soft_mask_dict(pdev, pparams);
         if (code < 0)
             return code;
         code = pdf_open_page(pdev, PDF_IN_STREAM);
         if (code < 0)
             return code;
-        return pdf_begin_transparency_group(pgs, pdev, pparams);
+        return pdf_begin_transparency_group(pgs, pdev, pparams, 0);
     }
 }
 
@@ -342,6 +403,7 @@ static int
 pdf_end_transparency_mask(gs_gstate * pgs, gx_device_pdf * pdev,
                                 const gs_pdf14trans_params_t * pparams)
 {
+    pdev->smask_construction = false;
     if (pdev->image_mask_skip)
         pdev->image_mask_skip = false;
     else {
@@ -357,7 +419,13 @@ pdf_end_transparency_mask(gs_gstate * pgs, gx_device_pdf * pdev,
             return 0;
         /* We need to update the 'where_used' field, in case we substituted a resource */
         pres->where_used |= pdev->used_mask;
-        gs_sprintf(buf, "%ld 0 R", pdf_resource_id(pres));
+        gs_snprintf(buf, sizeof(buf), "%ld 0 R", pdf_resource_id(pres));
+        if (pdev->pres_soft_mask_dict == 0L) {
+            /* something went horribly wrong, we have an 'end' wihtout a matching 'begin'
+             * Give up, throw an error.
+             */
+            return_error(gs_error_undefined);
+        }
         code = cos_dict_put_c_key_string((cos_dict_t *)pdev->pres_soft_mask_dict->object,
                 "/G", (const byte *)buf, strlen(buf));
         if (code < 0)
@@ -374,6 +442,7 @@ pdf_end_transparency_mask(gs_gstate * pgs, gx_device_pdf * pdev,
          * the FormDepth ourselves.
          */
         pdev->FormDepth--;
+        pdev->PatternsSinceForm = 0;
     }
     return 0;
 }
@@ -386,7 +455,7 @@ pdf_set_blend_params(gs_gstate * pgs, gx_device_pdf * dev,
 }
 
 int
-gdev_pdf_create_compositor(gx_device *dev,
+gdev_pdf_composite(gx_device *dev,
     gx_device **pcdev, const gs_composite_t *pct,
     gs_gstate *pgs, gs_memory_t *memory, gx_device *cdev)
 {
@@ -406,10 +475,16 @@ gdev_pdf_create_compositor(gx_device *dev,
                 return 0;
             case PDF14_ABORT_DEVICE:
                 return 0;
+            case PDF14_BEGIN_TRANS_PAGE_GROUP:
+                return pdf_begin_transparency_group(pgs, pdev, params, 1);
             case PDF14_BEGIN_TRANS_GROUP:
-                return pdf_begin_transparency_group(pgs, pdev, params);
+                return pdf_begin_transparency_group(pgs, pdev, params, 0);
             case PDF14_END_TRANS_GROUP:
                 return pdf_end_transparency_group(pgs, pdev);
+            case PDF14_BEGIN_TRANS_TEXT_GROUP:
+                return 0;
+            case PDF14_END_TRANS_TEXT_GROUP:
+                return 0;
             case PDF14_BEGIN_TRANS_MASK:
                 return pdf_begin_transparency_mask(pgs, pdev, params);
             case PDF14_END_TRANS_MASK:
@@ -430,7 +505,7 @@ gdev_pdf_create_compositor(gx_device *dev,
         }
         return 0;
     }
-    return psdf_create_compositor(dev, pcdev, pct, pgs, memory, cdev);
+    return psdf_composite(dev, pcdev, pct, pgs, memory, cdev);
 }
 
 /* We're not sure why the folllowing device methods are never called.
