@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2019 Artifex Software, Inc.
+/* Copyright (C) 2001-2024 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
@@ -29,6 +29,7 @@
 #include "gdevpdfo.h"
 #include "gdevpdtb.h"
 #include "gdevpdtf.h"
+#include "gdevpdtd.h"
 #include "smd5.h"
 #include "gxfcache.h"   /* for gs_purge_font_from_char_caches_completely */
 /*
@@ -46,8 +47,6 @@ private_st_pdf_base_font();
 gs_private_st_basic(st_pdf_base_font, pdf_base_font_t, "pdf_base_font_t",\
                     pdf_base_font_ptrs, pdf_base_font_data);
 
-#define SUBSET_PREFIX_SIZE 7	/* XXXXXX+ */
-
 typedef struct pdf_base14_font_info_s {
     const char *urwname;
     const char *stdname;
@@ -315,8 +314,10 @@ pdf_base_font_alloc(gx_device_pdf *pdev, pdf_base_font_t **ppbfont,
             complete = copied;
         else {
             code = gs_copy_font((gs_font *)font, &font->FontMatrix, mem, &complete, -1);
-            if (code < 0)
+            if (code < 0) {
+                gs_free_copied_font(copied);
                 goto fail;
+            }
         }
         code = gs_copy_font_complete((gs_font *)font, complete);
         if (code < 0 && pbfont->do_subset == DO_SUBSET_NO) {
@@ -328,6 +329,8 @@ pdf_base_font_alloc(gx_device_pdf *pdev, pdf_base_font_t **ppbfont,
             emprintf1(pdev->memory,
                       "Can't embed the complete font %s due to font error.\n",
                       buf);
+            gs_free_copied_font(copied);
+            copied = NULL;
             goto fail;
         }
         if (code < 0) {
@@ -348,7 +351,18 @@ pdf_base_font_alloc(gx_device_pdf *pdev, pdf_base_font_t **ppbfont,
      * this is required for PCL, see 'pdf_free_pdf_font_cache' in gdevpdf.c
      * for further details.
      */
+    pdev->pdf_font_dir->global_glyph_code = font->dir->global_glyph_code;
+
     pbfont->copied->dir = pbfont->complete->dir = pdev->pdf_font_dir;
+
+    if (pbfont->copied->FontType == ft_CID_encrypted) {
+        gs_font_cid0 *copied0 = (gs_font_cid0 *)pbfont->copied;
+        int i;
+        for (i = 0; i < copied0->cidata.FDArray_size; ++i) {
+            ((gs_font *)copied0->cidata.FDArray[i])->dir = pdev->pdf_font_dir;
+        }
+    }
+
     pbfont->is_standard = is_standard;
     if (pfname->size > 0) {
         font_name.data = pfname->chars;
@@ -359,7 +373,7 @@ pdf_base_font_alloc(gx_device_pdf *pdev, pdf_base_font_t **ppbfont,
             font_name.size -= SUBSET_PREFIX_SIZE;
         }
     } else {
-        gs_sprintf(fnbuf, ".F%lx", (ulong)copied);
+        gs_snprintf(fnbuf, sizeof(fnbuf), "Anonymous");
         font_name.data = (byte *)fnbuf;
         font_name.size = strlen(fnbuf);
     }
@@ -372,7 +386,7 @@ pdf_base_font_alloc(gx_device_pdf *pdev, pdf_base_font_t **ppbfont,
     *ppbfont = pbfont;
     return 0;
  fail:
-    gs_free_object(mem, pbfont, "pdf_base_font_alloc");
+    pdf_base_font_free(pdev, pbfont);
     return code;
 }
 
@@ -412,6 +426,12 @@ pdf_base_font_is_subset(const pdf_base_font_t *pbfont)
 void
 pdf_base_font_drop_complete(pdf_base_font_t *pbfont)
 {
+    /* gs_font is a subset of gs_font_base and we only want to
+     * free the members which are common to both, so this cast is
+     * (at the time of writing) safe.
+     */
+    if (pbfont->complete != pbfont->copied)
+        gs_free_copied_font((gs_font *)pbfont->complete);
     pbfont->complete = NULL;
 }
 
@@ -523,7 +543,7 @@ pdf_write_FontFile_entry(gx_device_pdf *pdev, pdf_base_font_t *pbfont)
  * Adjust font name for Acrobat Reader 3.
  */
 static int
-pdf_adjust_font_name(gx_device_pdf *pdev, long id, pdf_base_font_t *pbfont)
+pdf_adjust_font_name(gx_device_pdf *pdev, int64_t id, pdf_base_font_t *pbfont)
 {
     /*
      * In contradiction with previous version of pdfwrite,
@@ -536,7 +556,7 @@ pdf_adjust_font_name(gx_device_pdf *pdev, long id, pdf_base_font_t *pbfont)
     byte *chars = (byte *)pbfont->font_name.data; /* break 'const' */
     byte *data;
     uint size = pbfont->font_name.size;
-    char suffix[sizeof(long) * 2 + 2];
+    char suffix[sizeof(int64_t) * 2 + 2];
     uint suffix_size;
 
 #define SUFFIX_CHAR '~'
@@ -555,7 +575,7 @@ pdf_adjust_font_name(gx_device_pdf *pdev, long id, pdf_base_font_t *pbfont)
         size = i + 1;
     }
     /* Create a unique name. */
-    gs_sprintf(suffix, "%c%lx", SUFFIX_CHAR, id);
+    gs_snprintf(suffix, sizeof(suffix), "%c%lx", SUFFIX_CHAR, id);
     suffix_size = strlen(suffix);
     data = gs_resize_string(pdev->pdf_memory, chars, size,
                                   size + suffix_size,
@@ -654,11 +674,18 @@ pdf_write_embedded_font(gx_device_pdf *pdev, pdf_base_font_t *pbfont, font_type
             /* Write the type 1 font with no converting to CFF. */
             int lengths[3];
 
-            code = psf_write_type1_font(writer.binary.strm,
+            if (pbfont->do_subset != DO_SUBSET_NO)
+                code = psf_write_type1_font(writer.binary.strm,
                                 (gs_font_type1 *)out_font,
                                 WRITE_TYPE1_WITH_LENIV | WRITE_TYPE1_EEXEC |
                                 WRITE_TYPE1_EEXEC_PAD | WRITE_TYPE1_ASCIIHEX,
                                 NULL, 0, &fnstr, lengths);
+            else
+                code = psf_write_type1_font(writer.binary.strm,
+                                (gs_font_type1 *)out_font,
+                                WRITE_TYPE1_WITH_LENIV | WRITE_TYPE1_EEXEC |
+                                WRITE_TYPE1_EEXEC_PAD | WRITE_TYPE1_ASCIIHEX | WRITE_TYPE1_XUID,
+                                NULL, 0, &fnstr, lengths);
             if (lengths[0] > 0) {
                 if (code < 0)
                     goto finish;
@@ -689,7 +716,8 @@ pdf_write_embedded_font(gx_device_pdf *pdev, pdf_base_font_t *pbfont, font_type
             code = psf_write_type2_font(writer.binary.strm,
                                         (gs_font_type1 *)out_font,
                                         TYPE2_OPTIONS |
-                            (pdev->CompatibilityLevel < 1.3 ? WRITE_TYPE2_AR3 : 0),
+                            (pdev->CompatibilityLevel < 1.3 ? WRITE_TYPE2_AR3 : 0) |
+                            (pbfont->do_subset == DO_SUBSET_NO ? WRITE_TYPE2_XUID : 0),
                                         NULL, 0, &fnstr, FontBBox);
         }
         goto finish;
@@ -718,10 +746,10 @@ pdf_write_embedded_font(gx_device_pdf *pdev, pdf_base_font_t *pbfont, font_type
         swrite_position_only(&poss);
         code = psf_write_truetype_font(&poss, pfont, options, NULL, 0, &fnstr);
         if (code < 0)
-            return code;
+            goto finish;
         code = cos_dict_put_c_key_int((cos_dict_t *)writer.pres->object, "/Length1", stell(&poss));
         if (code < 0)
-            return code;
+            goto finish;
         code = psf_write_truetype_font(writer.binary.strm, pfont,
                                        options, NULL, 0, &fnstr);
         goto finish;
@@ -799,7 +827,7 @@ pdf_write_CharSet(gx_device_pdf *pdev, pdf_base_font_t *pbfont)
  */
 int
 pdf_write_CIDSet(gx_device_pdf *pdev, pdf_base_font_t *pbfont,
-                 long *pcidset_id)
+                 int64_t *pcidset_id)
 {
     pdf_data_writer_t writer;
     int code;
