@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2012 Artifex Software, Inc.
+/* Copyright (C) 2001-2024 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  7 Mt. Lassen Drive - Suite A-134, San Rafael,
-   CA  94903, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
@@ -112,7 +112,7 @@ gs_public_st_basic_super(st_pdf_sub_font_descriptor,
 /* ---------------- Private ---------------- */
 
 /* Get the ID of font descriptor metrics. */
-static inline long
+static inline int64_t
 pdf_font_descriptor_common_id(const pdf_font_descriptor_common_t *pfdc)
 {
     return pdf_resource_id((const pdf_resource_t *)pfdc);
@@ -212,8 +212,6 @@ pdf_font_descriptor_alloc(gx_device_pdf *pdev, pdf_font_descriptor_t **ppfd,
                        "pdf_font_descriptor_alloc(base_font)");
         return code;
     }
-    memset(&pfd->common.values, 0,
-           sizeof(*pfd) - offset_of(pdf_font_descriptor_t, common.values));
     pfd->base_font = pbfont;
     pfd->FontType = font->FontType;
     pfd->embed = embed;
@@ -221,10 +219,8 @@ pdf_font_descriptor_alloc(gx_device_pdf *pdev, pdf_font_descriptor_t **ppfd,
     return 0;
 }
 
-int pdf_font_descriptor_free(gx_device_pdf *pdev, pdf_resource_t *pres)
+int pdf_base_font_free(gx_device_pdf *pdev, pdf_base_font_t *pbfont)
 {
-    pdf_font_descriptor_t *pfd = (pdf_font_descriptor_t *)pres;
-    pdf_base_font_t *pbfont = pfd->base_font;
     gs_font *copied = NULL, *complete = NULL;
 
     if (pbfont) {
@@ -239,14 +235,27 @@ int pdf_font_descriptor_free(gx_device_pdf *pdev, pdf_resource_t *pres)
         gs_free_copied_font(copied);
 
     if (pbfont) {
+        if (pbfont->CIDSet)
+            gs_free_object(pdev->pdf_memory, pbfont->CIDSet, "Free base font CIDSet from FontDescriptor)");
+
         if (pbfont->font_name.size) {
             gs_free_string(pdev->pdf_memory, pbfont->font_name.data, pbfont->font_name.size, "Free BaseFont FontName string");
             pbfont->font_name.data = (byte *)0L;
             pbfont->font_name.size = 0;
         }
-        gs_free_object(cos_object_memory(pres->object), pbfont, "Free base font from FontDescriptor)");
-        pfd->base_font = 0;
+        gs_free_object(pdev->pdf_memory, pbfont, "Free base font from FontDescriptor)");
     }
+    return 0;
+}
+
+int pdf_font_descriptor_free(gx_device_pdf *pdev, pdf_resource_t *pres)
+{
+    pdf_font_descriptor_t *pfd = (pdf_font_descriptor_t *)pres;
+    pdf_base_font_t *pbfont = pfd->base_font;
+
+    pdf_base_font_free(pdev, pbfont);
+    pfd->base_font = 0;
+
     if (pres->object) {
         gs_free_object(pdev->pdf_memory, pres->object, "free FontDescriptor object");
         pres->object = NULL;
@@ -255,13 +264,13 @@ int pdf_font_descriptor_free(gx_device_pdf *pdev, pdf_resource_t *pres)
 }
 
 /* Get the object ID of a FontDescriptor. */
-long
+int64_t
 pdf_font_descriptor_id(const pdf_font_descriptor_t *pfd)
 {
     return pdf_resource_id((const pdf_resource_t *)pfd);
 }
 
-long pdf_set_font_descriptor_usage(gx_device_pdf *pdev, int parent_id, const pdf_font_descriptor_t *pfd)
+int64_t pdf_set_font_descriptor_usage(gx_device_pdf *pdev, int parent_id, const pdf_font_descriptor_t *pfd)
 {
     int id = pdf_resource_id((const pdf_resource_t *)pfd);
 
@@ -421,6 +430,7 @@ pdf_compute_font_descriptor(gx_device_pdf *pdev, pdf_font_descriptor_t *pfd)
     case ft_PCL_user_defined:
     case ft_GL2_stick_user_defined:
     case ft_MicroType:
+    case ft_PDF_user_defined:
     case ft_user_defined:
         break;
         /* Other font types may use a non-standard (not 1000x1000) design grid
@@ -555,7 +565,7 @@ pdf_compute_font_descriptor(gx_device_pdf *pdev, pdf_font_descriptor_t *pfd)
     if (!(desc.Flags & FONT_IS_SYMBOLIC)) {
         desc.Flags |= FONT_IS_ADOBE_ROMAN; /* required if not symbolic */
         desc.XHeight = (int)x_height;
-        if (!small_present && (!pdev->PDFA != 0 || bfont->FontType != ft_TrueType))
+        if (!small_present && (pdev->PDFA == 0 || bfont->FontType != ft_TrueType))
             desc.Flags |= FONT_IS_ALL_CAPS;
         desc.CapHeight = cap_height;
         /*
@@ -606,10 +616,10 @@ pdf_compute_font_descriptor(gx_device_pdf *pdev, pdf_font_descriptor_t *pfd)
     desc.Descent = desc.FontBBox.p.y;
     if (!(desc.Flags & (FONT_IS_SYMBOLIC | FONT_IS_ALL_CAPS)) &&
         (small_descent > desc.Descent / 3 || desc.XHeight > small_height * 0.9) &&
-        (!pdev->PDFA != 0 || bfont->FontType != ft_TrueType)
+        (pdev->PDFA == 0 || bfont->FontType != ft_TrueType)
         )
         desc.Flags |= FONT_IS_SMALL_CAPS;
-    if (fixed_width > 0 && (!pdev->PDFA != 0 || bfont->FontType != ft_TrueType)) {
+    if (fixed_width > 0 && (pdev->PDFA == 0 || bfont->FontType != ft_TrueType)) {
         desc.Flags |= FONT_IS_FIXED_WIDTH;
         desc.AvgWidth = desc.MaxWidth = desc.MissingWidth = fixed_width;
     }
@@ -652,7 +662,7 @@ pdf_write_FontDescriptor(gx_device_pdf *pdev, pdf_resource_t *pres)
 {
     pdf_font_descriptor_t *pfd = (pdf_font_descriptor_t *)pres;
     font_type ftype = pfd->FontType;
-    long cidset_id = 0;
+    int64_t cidset_id = 0;
     int code = 0;
     stream *s;
 
@@ -705,7 +715,8 @@ pdf_write_FontDescriptor(gx_device_pdf *pdev, pdf_resource_t *pres)
     if (cidset_id != 0)
         pprintld1(s, "/CIDSet %ld 0 R\n", cidset_id);
     else if (pdf_do_subset_font(pdev, pfd->base_font, pfd->common.rid) &&
-             (ftype == ft_encrypted || ftype == ft_encrypted2)
+             (ftype == ft_encrypted || ftype == ft_encrypted2) &&
+             pdev->CompatibilityLevel <= 1.7
              ) {
         stream_puts(s, "/CharSet");
         code = pdf_write_CharSet(pdev, pfd->base_font);
