@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2020 Artifex Software, Inc.
+/* Copyright (C) 2001-2024 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
@@ -117,13 +117,30 @@ typedef int (*pdf_font_write_contents_proc_t)
  */
 typedef struct pdf_encoding_element_s {
     gs_glyph glyph;
-    gs_const_string str;
+    byte *data;
+    uint size;
     bool is_difference;		/* true if must be written in Differences */
 } pdf_encoding_element_t;
+
+static inline int pdf_copy_string_to_encoding(gx_device_pdf *pdev, gs_const_string *gnstr, pdf_encoding_element_t *pet)
+{
+    byte *p = NULL;
+
+    p = gs_alloc_bytes(pdev->pdf_memory->non_gc_memory, gnstr->size, "pdf_copy_string_to_encoding");
+    if (p == NULL)
+        return_error(gs_error_VMerror);
+    memcpy(p, gnstr->data, gnstr->size);
+    if (pet->data != NULL)
+        gs_free_object(pdev->pdf_memory->non_gc_memory, pet->data, "pdf_copy_string_to_encoding free existing glyph name");
+    pet->data = p;
+    pet->size = gnstr->size;
+    return 0;
+}
+
 #define private_st_pdf_encoding1() /* gdevpdtf.c */\
   gs_private_st_const_strings1(st_pdf_encoding1,\
     pdf_encoding_element_t, "pdf_encoding_element_t",\
-    pdf_encoding1_enum_ptrs, pdf_encoding1_reloc_ptrs, str)
+    pdf_encoding1_enum_ptrs, pdf_encoding1_reloc_ptrs, bah)
 #define private_st_pdf_encoding_element() /* gdevpdtf.c */\
   gs_private_st_element(st_pdf_encoding_element, pdf_encoding_element_t,\
     "pdf_encoding_element_t[]", pdf_encoding_elt_enum_ptrs,\
@@ -185,7 +202,7 @@ typedef struct {
 struct pdf_font_resource_s {
     pdf_resource_common(pdf_font_resource_t);
     font_type FontType;		/* copied from font, if any */
-    long XUID;
+    int64_t XUID;
     pdf_font_write_contents_proc_t write_contents;
     gs_string BaseFont;		/* (not used for Type 3) */
     pdf_font_descriptor_t *FontDescriptor; /* (not used for Type 0, Type 3, */
@@ -224,11 +241,11 @@ struct pdf_font_resource_s {
              */
             char Encoding_name[max( /* standard name or <id> 0 R */
                       17,	/* /UniJIS-UCS2-HW-H */
-                      sizeof(long) * 8 / 3 + 1 + 4 /* <id> 0 R */
+                      sizeof(int64_t) * 8 / 3 + 1 + 4 /* <id> 0 R */
                       ) + 1	/* \0 terminator */
             ];
-            gs_const_string CMapName; /* copied from the original CMap, */
-                                /* or references the table of standard names */
+            byte *CMapName_data;
+            uint CMapName_size;
             uint font_index;	/* The index of the descendent font in the source CMap. */
             bool cmap_is_standard;
             int WMode;		/* of CMap */
@@ -238,7 +255,7 @@ struct pdf_font_resource_s {
         struct /*cidfont*/ {
 
             /* [D]W[2] is Widths. */
-            long CIDSystemInfo_id; /* (written when font is allocated) */
+            int64_t CIDSystemInfo_id; /* (written when font is allocated) */
             ushort *CIDToGIDMap; /* (CIDFontType 2 only) [count] */
             unsigned int CIDToGIDMapLength;
             gs_id glyphshow_font_id;
@@ -392,7 +409,7 @@ int font_resource_free(gx_device_pdf *pdev, pdf_font_resource_t *pdfont);
 
 /* Resize font resource arrays. */
 int pdf_resize_resource_arrays(gx_device_pdf *pdev, pdf_font_resource_t *pfres,
-        int chars_count);
+        uint chars_count);
 
 /*
  * Return the (copied, subset or complete) font associated with a font resource.
@@ -409,7 +426,7 @@ gs_font_base *pdf_font_resource_font(const pdf_font_resource_t *pdfont, bool com
  */
 pdf_font_embed_t pdf_font_embed_status(gx_device_pdf *pdev, gs_font *font,
                                        int *pindex,
-                                       pdf_char_glyph_pair_t *pairs, int num_glyphs);
+                                       pdf_char_glyph_pair_t *pairs, int num_glyphs, font_type *orig_type);
 
 /*
  * Compute the BaseFont of a font according to the algorithm described
