@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2019 Artifex Software, Inc.
+/* Copyright (C) 2001-2024 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
@@ -153,7 +153,7 @@ extern const stream_template s_Average_template;
 typedef struct stream_Bicubic_state_s {
     stream_Downsample_state_common;
     int y_in;
-    ulong l_size, d_size, d_len;
+    uint64_t l_size, d_size, d_len;
     byte *data;
 } stream_Bicubic_state;
 
@@ -169,10 +169,10 @@ typedef struct stream_compr_chooser_state_s {
     uint choice;
     uint width, height, depth, bits_per_sample;
     uint samples_count, bits_left;
-    ulong packed_data;
+    uint64_t packed_data;
     byte *sample;
-    ulong upper_plateaus, lower_plateaus;
-    ulong gradients;
+    uint64_t upper_plateaus, lower_plateaus;
+    uint64_t gradients;
 } stream_compr_chooser_state;
 
 #define private_st_compr_chooser_state()	/* in gdevpsds.c */\
