@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2019 Artifex Software, Inc.
+/* Copyright (C) 2001-2023 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
@@ -640,8 +640,8 @@ void main(void)
     argv[1] = ddevice;
     argv[2] = dformat;
 
-    gs_sprintf(ddevice, "-sDEVICE=display");
-    gs_sprintf(dformat, "-dDisplayFormat=%d", display_format);
+    gs_snprintf(ddevice, sizeof(ddevice), "-sDEVICE=display");
+    gs_snprintf(dformat, sizeof(dformat), "-dDisplayFormat=%d", display_format);
 
     /* Run Ghostscript */
     if (gsapi_new_instance(&instance, NULL) < 0)
